/*
 * Handle unaligned accesses by emulation.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1996 by Ralf Baechle
 *
 * This file contains exception handler for address error exception with the
 * special capability to execute faulting instructions in software.  The
 * handler does not try to handle the case when the program counter points
 * to an address not aligned to a word boundary.
 *
 * Putting data to unaligned addresses is a bad practice even on Intel where
 * only the performance is affected.  Much worse is that such code is non-
 * portable.  Due to several programs that die on MIPS due to alignment
 * problems I decieded to implement this handler anyway though I originally
 * didn't intend to do this at all for user code.
 *
 * For now I enable fixing of address errors by default to make life easier.
 * I however intend to disable this somewhen in the future when the alignment
 * problems with user programs have been fixed.  For programmers this is the
 * right way to go.
 *
 * Fixing address errors is a per process option.  The option is inherited
 * across fork(2) and execve(2) calls.  If you really want to use the
 * option in your user programs - I discourage the use of the software
 * emulation strongly - use the following code in your userland stuff:
 *
 * #include <sys/sysmips.h>
 *
 * ...
 * sysmips(MIPS_FIXADE, x);
 * ...
 *
 * The parameter x is 0 for disabeling software emulation.  Set bit 0 for
 * enabeling software emulation and bit 1 for enabeling printing debug
 * messages into syslog to aid finding address errors in programs.
 *
 * The logging feature is a addition over RISC/os and IRIX where only the
 * values 0 and 1 are acceptable values for x.  I'll probably remove this
 * hack later on.
 *
 * Below a little program to play around with this feature.
 *
 * #include <stdio.h>
 * #include <asm/sysmips.h>
 * 
 * struct foo {
 *         unsigned char bar[8];
 * };
 *
 * main(int argc, char *argv[])
 * {
 *         struct foo x = {0, 1, 2, 3, 4, 5, 6, 7};
 *         unsigned int *p = (unsigned int *) (x.bar + 3);
 *         int i;
 *
 *         if (argc > 1)
 *                 sysmips(MIPS_FIXADE, atoi(argv[1]));
 *
 *         printf("*p = %08lx\n", *p);
 *
 *         *p = 0xdeadface;
 *
 *         for(i = 0; i <= 7; i++)
 *         printf("%02x ", x.bar[i]);
 *         printf("\n");
 * }
 *
 * Until I've written the code to handle branch delay slots it may happen
 * that the kernel receives an ades/adel instruction from an insn in a
 * branch delay slot but is unable to handle this case.  The kernel knows
 * this fact and therefore will kill the process.  For most code you can
 * fix this temporarily by compiling with flags -fno-delayed-branch -Wa,-O0.
 *
 * Coprozessor loads are not supported; I think this case is unimportant
 * in the practice.
 *
 * TODO: Handle ndc (attempted store to doubleword in uncached memory)
 *       exception for the R6000.
 */
#include <linux/mm.h>
#include <linux/signal.h>
#include <asm/inst.h>
#include <asm/segment.h>
#include <asm/unaligned.h>

/*
 * Don't inline the functions for better debugging.
 */
#define __inline__

extern int compute_return_epc(struct pt_regs *regs, char *exc);
/*
 * Damned ...  There is just too much bad code with that Intel desease in
 * it out there.   So we really need to support this :-(
 */
static __inline__ __uregister_t
load_mem(unsigned long addr, unsigned int size)
{
	switch(size) {
	case 2: return __ulh((unsigned short *) addr);
	case 4: return __ulw((unsigned int *) addr);
#if (_MIPS_ISA == _MIPS_ISA_MIPS3) || (_MIPS_ISA == _MIPS_ISA_MIPS4)
	case 8: return __uld((unsigned long long *) addr);
#endif
	}

	/* Unreached */
	return 0;
}

static __inline__ void
store_mem(__uregister_t val, unsigned long addr, unsigned int size)
{
	switch(size) {
	case 2: __ush(val, (unsigned short *) addr);
		return;
	case 4: __usw(val, (unsigned int *) addr);
		return;
#if (_MIPS_ISA == _MIPS_ISA_MIPS3) || (_MIPS_ISA == _MIPS_ISA_MIPS4)
	case 8: __usd(val, (unsigned long long *) addr);
		return;
#endif
	}
}

static __inline__ void
emulate_load_store_insn(struct pt_regs *regs, unsigned long addr, unsigned long pc)
{
	union mips_instruction insn;
	int errno, size, signop;
	__register_t value;

	/*
	 * Assumes we're not in a branch delay slot.
	 */
	insn.word = get_user((unsigned int *)pc);
	switch (insn.i_format.opcode) {
	/*
	 * These are instruction that a compiler doesn't generate.  We
	 * can assume therefore that the code is MIPS-aware and
	 * really buggy.  Emulating these instructions would break the
	 * semantics anyway.
	 */
	case ll_op:
	case lld_op:
	case sc_op:
	case scd_op:

	/*
	 * For these instructions the only way to create an address
	 * error is an attempted access to kernel/supervisor address
	 * space.
	 */
	case ldl_op:
	case ldr_op:
	case lwl_op:
	case lwr_op:
	case sdl_op:
	case sdr_op:
	case swl_op:
	case swr_op:
	case lb_op:
	case lbu_op:
	case sb_op:
		force_sig(SIGBUS, current);
		return;

	/*
	 * The remaining opcodes are the ones that are really of interrest.
	 */
	case lh_op:	size = 2; signop = 1; goto do_reg_load;
	case lhu_op:	size = 2; signop = 0; goto do_reg_load;
	case lw_op:	size = 4; signop = 1; goto do_reg_load;
	case lwu_op:	size = 4; signop = 0; goto do_reg_load;
	case ld_op:	size = 8; signop = 1;
do_reg_load:
		errno = verify_area(VERIFY_READ, (void *)addr, size);
		if (errno < 0) {
			force_sig(SIGSEGV, current);
			return;
		}
		value = load_mem(addr, size);
		if (!signop) {
			/* This is an unsigned load */
			if (size == 2) value &= 0xffffUL;
			else if (size == 4) value &= 0xffffffffUL;
		}
		regs->regs[insn.i_format.rt] = value;
		regs->regs[insn.i_format.rt] = value;
		return;
	case sh_op:	size = 2; goto do_reg_store;
	case sw_op:	size = 4; goto do_reg_store;
	case sd_op:	size = 8;
do_reg_store:
		errno = verify_area(VERIFY_WRITE, (void *)addr, size);
		if (errno < 0) {
			force_sig(SIGSEGV, current);
			return;
		}
		store_mem(regs->regs[insn.i_format.rt], addr, size);
		return;

	case lwc1_op:
	case ldc1_op:
	case swc1_op:
	case sdc1_op:
		/*
		 * I herewith declare: this does not happen.  So send SIGBUS.
		 */
		force_sig(SIGBUS, current);
		return;
	case lwc2_op:
	case ldc2_op:
	case swc2_op:
	case sdc2_op:
		/*
		 * These are the coprozessor 2 load/stores.  The current
		 * implementations don't use cp2 and cp2 should always be
		 * disabled in c0_status.  So send SIGILL.
		 */
	default:
		/*
		 * Pheeee...  We encountered an yet unknown instruction ...
		 */
		force_sig(SIGILL, current);
	}
}

static __inline__ void
fix_ade(struct pt_regs *regs, unsigned long pc, char *exc)
{
	/*
	 * Did we catch a fault trying to load an instruction?
	 */
	if (regs->cp0_badvaddr == pc) {
		/*
		 * This fault occured during an attempt to load an instruction.
		 */
		force_sig(SIGBUS, current);
	}

	/*
	 * Ok, this wasn't a failed instruction load.  The CPU was capable of
	 * reading the instruction and faulted after this.  So we don't need
	 * to verify_area the address of the instrucion.  We still don't
	 * know whether the address used was legal and therefore need to do
	 * verify_area().  The CPU already did the checking for legal
	 * instructions for us, so we don't need to do this.
	 */
	emulate_load_store_insn(regs, regs->cp0_badvaddr, pc);
}

#define is_kernel_address(x) ((x) & 0x80000000)

asmlinkage void
do_ade(struct pt_regs *regs)
{
	__register_t logpc, pc = regs->cp0_epc;
	__register_t badvaddr = regs->cp0_badvaddr;
	char *adels;

	adels = (((regs->cp0_cause & CAUSEF_EXCCODE) >> CAUSEB_EXCCODE) == 4) ? "adel" : "ades";
	/*
	 * I think there are no legal unaligned kernel accesses to the kernel
	 * space; we should fix them anyway.  So let's do a little panic ...
	 */
	if (is_kernel_address(badvaddr) && !user_mode(regs)) {
		show_regs(regs);
#ifdef __mips64
		panic("Caught %s exception in kernel mode accessing %016Lx.", adels, badvaddr);
#else
		panic("Caught %s exception in kernel mode accessing %08lx.", adels, badvaddr);
#endif
	}
	if (current->tss.mflags & MF_LOGADE) {
		logpc = pc;
		if (regs->cp0_cause & CAUSEF_BD)
			logpc += 4;
#ifdef __mips64
		printk(KERN_DEBUG
		       "Caught %s in '%s' at 0x%016Lx accessing 0x%016Lx.\n",
		       adels, current->comm, logpc, regs->cp0_badvaddr);
#else
		printk(KERN_DEBUG
		       "Caught %s in '%s' at 0x%08lx accessing 0x%08lx.\n",
		       adels, current->comm, logpc, regs->cp0_badvaddr);
#endif
	}
	if (compute_return_epc(regs, adels))
		return;
	if(current->tss.mflags & MF_FIXADE) {
		fix_ade(regs, pc, adels);
		return;
	}
#ifdef CONF_DEBUG_EXCEPTIONS
	show_regs(regs);
#endif
	force_sig(SIGBUS, current);
}
